@@ -1,21 +1,21 @@
 /* Subroutines used for code generation on the Argonaut ARC cpu.
-   Copyright (C) 1994, 1995, 1997, 1998, 1999, 2000, 2001, 2002
+   Copyright (C) 1994, 1995, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004
    Free Software Foundation, Inc.
 
-This file is part of GNU CC.
+This file is part of GCC.
 
-GNU CC is free software; you can redistribute it and/or modify
+GCC is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.
 
-GNU CC is distributed in the hope that it will be useful,
+GCC is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
-along with GNU CC; see the file COPYING.  If not, write to
+along with GCC; see the file COPYING.  If not, write to
 the Free Software Foundation, 59 Temple Place - Suite 330,
 Boston, MA 02111-1307, USA.  */
 
@@ -23,6 +23,8 @@ Boston, MA 02111-1307, USA.  */
 
 #include "config.h"
 #include "system.h"
+#include "coretypes.h"
+#include "tm.h"
 #include "tree.h"
 #include "rtl.h"
 #include "regs.h"
@@ -84,15 +86,24 @@ static int arc_ccfsm_target_label;
    arc_print_operand.  */
 static int last_insn_set_cc_p;
 static int current_insn_set_cc_p;
-static void record_cc_ref PARAMS ((rtx));
-static void arc_init_reg_tables PARAMS ((void));
-static int get_arc_condition_code PARAMS ((rtx));
+static void record_cc_ref (rtx);
+static void arc_init_reg_tables (void);
+static int get_arc_condition_code (rtx);
 const struct attribute_spec arc_attribute_table[];
-static tree arc_handle_interrupt_attribute PARAMS ((tree *, tree, tree, int, bool *));
-static bool arc_assemble_integer PARAMS ((rtx, unsigned int, int));
-static void arc_output_function_prologue PARAMS ((FILE *, HOST_WIDE_INT));
-static void arc_output_function_epilogue PARAMS ((FILE *, HOST_WIDE_INT));
-static void arc_encode_section_info PARAMS ((tree, int));
+static tree arc_handle_interrupt_attribute (tree *, tree, tree, int, bool *);
+static bool arc_assemble_integer (rtx, unsigned int, int);
+static void arc_output_function_prologue (FILE *, HOST_WIDE_INT);
+static void arc_output_function_epilogue (FILE *, HOST_WIDE_INT);
+static void arc_file_start (void);
+static void arc_internal_label (FILE *, const char *, unsigned long);
+static void arc_setup_incoming_varargs (CUMULATIVE_ARGS *, enum machine_mode,
+					tree, int *, int);
+static bool arc_rtx_costs (rtx, int, int, int *);
+static int arc_address_cost (rtx);
+static void arc_external_libcall (rtx);
+static bool arc_return_in_memory (tree, tree);
+static bool arc_pass_by_reference (CUMULATIVE_ARGS *, enum machine_mode,
+				   tree, bool);
 
 /* Initialize the GCC target structure.  */
 #undef TARGET_ASM_ALIGNED_HI_OP
@@ -106,17 +117,43 @@ static void arc_encode_section_info PARAMS ((tree, int));
 #define TARGET_ASM_FUNCTION_PROLOGUE arc_output_function_prologue
 #undef TARGET_ASM_FUNCTION_EPILOGUE
 #define TARGET_ASM_FUNCTION_EPILOGUE arc_output_function_epilogue
+#undef TARGET_ASM_FILE_START
+#define TARGET_ASM_FILE_START arc_file_start
 #undef TARGET_ATTRIBUTE_TABLE
 #define TARGET_ATTRIBUTE_TABLE arc_attribute_table
-#undef TARGET_ENCODE_SECTION_INFO
-#define TARGET_ENCODE_SECTION_INFO arc_encode_section_info
+#undef TARGET_ASM_INTERNAL_LABEL
+#define TARGET_ASM_INTERNAL_LABEL arc_internal_label
+#undef TARGET_ASM_EXTERNAL_LIBCALL
+#define TARGET_ASM_EXTERNAL_LIBCALL arc_external_libcall
+
+#undef TARGET_RTX_COSTS
+#define TARGET_RTX_COSTS arc_rtx_costs
+#undef TARGET_ADDRESS_COST
+#define TARGET_ADDRESS_COST arc_address_cost
+
+#undef TARGET_PROMOTE_FUNCTION_ARGS
+#define TARGET_PROMOTE_FUNCTION_ARGS hook_bool_tree_true
+#undef TARGET_PROMOTE_FUNCTION_RETURN
+#define TARGET_PROMOTE_FUNCTION_RETURN hook_bool_tree_true
+#undef TARGET_PROMOTE_PROTOTYPES
+#define TARGET_PROMOTE_PROTOTYPES hook_bool_tree_true
+
+#undef TARGET_RETURN_IN_MEMORY
+#define TARGET_RETURN_IN_MEMORY arc_return_in_memory
+#undef TARGET_PASS_BY_REFERENCE
+#define TARGET_PASS_BY_REFERENCE arc_pass_by_reference
+#undef TARGET_CALLEE_COPIES
+#define TARGET_CALLEE_COPIES hook_bool_CUMULATIVE_ARGS_mode_tree_bool_true
+
+#undef TARGET_SETUP_INCOMING_VARARGS
+#define TARGET_SETUP_INCOMING_VARARGS arc_setup_incoming_varargs
 
 struct gcc_target targetm = TARGET_INITIALIZER;
 
 /* Called by OVERRIDE_OPTIONS to initialize various things.  */
 
 void
-arc_init ()
+arc_init (void)
 {
   char *tmp;
   
@@ -171,8 +208,7 @@ static const char *const arc_condition_codes[] =
    `(eq (...) (...))'.  */
 
 static int
-get_arc_condition_code (comparison)
-     rtx comparison;
+get_arc_condition_code (rtx comparison)
 {
   switch (GET_CODE (comparison))
     {
@@ -196,9 +232,9 @@ get_arc_condition_code (comparison)
    return the mode to be used for the comparison.  */
 
 enum machine_mode
-arc_select_cc_mode (op, x, y)
-     enum rtx_code op;
-     rtx x, y ATTRIBUTE_UNUSED;
+arc_select_cc_mode (enum rtx_code op,
+	            rtx x ATTRIBUTE_UNUSED,
+                    rtx y ATTRIBUTE_UNUSED)
 {
   switch (op)
     {
@@ -272,7 +308,7 @@ unsigned int arc_mode_class [NUM_MACHINE_MODES];
 enum reg_class arc_regno_reg_class[FIRST_PSEUDO_REGISTER];
 
 static void
-arc_init_reg_tables ()
+arc_init_reg_tables (void)
 {
   int i;
 
@@ -308,13 +344,10 @@ arc_init_reg_tables ()
 	    arc_mode_class[i] = 0;
 	  break;
 	case MODE_CC:
+	  arc_mode_class[i] = 1 << (int) C_MODE;
+	  break;
 	default:
-	  /* mode_class hasn't been initialized yet for EXTRA_CC_MODES, so
-	     we must explicitly check for them here.  */
-	  if (i == (int) CCmode || i == (int) CCZNmode || i == (int) CCZNCmode)
-	    arc_mode_class[i] = 1 << (int) C_MODE;
-	  else
-	    arc_mode_class[i] = 0;
+	  arc_mode_class[i] = 0;
 	  break;
 	}
     }
@@ -348,25 +381,24 @@ const struct attribute_spec arc_attribute_table[] =
 /* Handle an "interrupt" attribute; arguments as in
    struct attribute_spec.handler.  */
 static tree
-arc_handle_interrupt_attribute (node, name, args, flags, no_add_attrs)
-     tree *node ATTRIBUTE_UNUSED;
-     tree name;
-     tree args;
-     int flags ATTRIBUTE_UNUSED;
-     bool *no_add_attrs;
+arc_handle_interrupt_attribute (tree *node ATTRIBUTE_UNUSED,
+                                tree name,
+                                tree args,
+                                int flags ATTRIBUTE_UNUSED,
+                                bool *no_add_attrs)
 {
   tree value = TREE_VALUE (args);
 
   if (TREE_CODE (value) != STRING_CST)
     {
-      warning ("argument of `%s' attribute is not a string constant",
+      warning ("argument of %qs attribute is not a string constant",
 	       IDENTIFIER_POINTER (name));
       *no_add_attrs = true;
     }
   else if (strcmp (TREE_STRING_POINTER (value), "ilink1")
 	   && strcmp (TREE_STRING_POINTER (value), "ilink2"))
     {
-      warning ("argument of `%s' attribute is not \"ilink1\" or \"ilink2\"",
+      warning ("argument of %qs attribute is not \"ilink1\" or \"ilink2\"",
 	       IDENTIFIER_POINTER (name));
       *no_add_attrs = true;
     }
@@ -378,9 +410,7 @@ arc_handle_interrupt_attribute (node, name, args, flags, no_add_attrs)
 /* Acceptable arguments to the call insn.  */
 
 int
-call_address_operand (op, mode)
-     rtx op;
-     enum machine_mode mode;
+call_address_operand (rtx op, enum machine_mode mode)
 {
   return (symbolic_operand (op, mode)
 	  || (GET_CODE (op) == CONST_INT && LEGITIMATE_CONSTANT_P (op))
@@ -388,9 +418,7 @@ call_address_operand (op, mode)
 }
 
 int
-call_operand (op, mode)
-     rtx op;
-     enum machine_mode mode;
+call_operand (rtx op, enum machine_mode mode)
 {
   if (GET_CODE (op) != MEM)
     return 0;
@@ -401,9 +429,7 @@ call_operand (op, mode)
 /* Returns 1 if OP is a symbol reference.  */
 
 int
-symbolic_operand (op, mode)
-     rtx op;
-     enum machine_mode mode ATTRIBUTE_UNUSED;
+symbolic_operand (rtx op, enum machine_mode mode ATTRIBUTE_UNUSED)
 {
   switch (GET_CODE (op))
     {
@@ -420,9 +446,7 @@ symbolic_operand (op, mode)
    operand of mode MODE.  */
 
 int
-symbolic_memory_operand (op, mode)
-     rtx op;
-     enum machine_mode mode ATTRIBUTE_UNUSED;
+symbolic_memory_operand (rtx op, enum machine_mode mode ATTRIBUTE_UNUSED)
 {
   if (GET_CODE (op) == SUBREG)
     op = SUBREG_REG (op);
@@ -436,9 +460,7 @@ symbolic_memory_operand (op, mode)
 /* Return true if OP is a short immediate (shimm) value.  */
 
 int
-short_immediate_operand (op, mode)
-     rtx op;
-     enum machine_mode mode ATTRIBUTE_UNUSED;
+short_immediate_operand (rtx op, enum machine_mode mode ATTRIBUTE_UNUSED)
 {
   if (GET_CODE (op) != CONST_INT)
     return 0;
@@ -449,9 +471,7 @@ short_immediate_operand (op, mode)
    This is currently only used when calculating length attributes.  */
 
 int
-long_immediate_operand (op, mode)
-     rtx op;
-     enum machine_mode mode ATTRIBUTE_UNUSED;
+long_immediate_operand (rtx op, enum machine_mode mode ATTRIBUTE_UNUSED)
 {
   switch (GET_CODE (op))
     {
@@ -479,9 +499,8 @@ long_immediate_operand (op, mode)
    This is currently only used when calculating length attributes.  */
 
 int
-long_immediate_loadstore_operand (op, mode)
-     rtx op;
-     enum machine_mode mode ATTRIBUTE_UNUSED;
+long_immediate_loadstore_operand (rtx op,
+                                  enum machine_mode mode ATTRIBUTE_UNUSED)
 {
   if (GET_CODE (op) != MEM)
     return 0;
@@ -521,9 +540,7 @@ long_immediate_loadstore_operand (op, mode)
    move source.  */
 
 int
-move_src_operand (op, mode)
-     rtx op;
-     enum machine_mode mode;
+move_src_operand (rtx op, enum machine_mode mode)
 {
   switch (GET_CODE (op))
     {
@@ -563,9 +580,7 @@ move_src_operand (op, mode)
    move source.  */
 
 int
-move_double_src_operand (op, mode)
-     rtx op;
-     enum machine_mode mode;
+move_double_src_operand (rtx op, enum machine_mode mode)
 {
   switch (GET_CODE (op))
     {
@@ -595,9 +610,7 @@ move_double_src_operand (op, mode)
 /* Return true if OP is an acceptable argument for a move destination.  */
 
 int
-move_dest_operand (op, mode)
-     rtx op;
-     enum machine_mode mode;
+move_dest_operand (rtx op, enum machine_mode mode)
 {
   switch (GET_CODE (op))
     {
@@ -620,9 +633,7 @@ move_dest_operand (op, mode)
 /* Return true if OP is valid load with update operand.  */
 
 int
-load_update_operand (op, mode)
-     rtx op;
-     enum machine_mode mode;
+load_update_operand (rtx op, enum machine_mode mode)
 {
   if (GET_CODE (op) != MEM
       || GET_MODE (op) != mode)
@@ -639,9 +650,7 @@ load_update_operand (op, mode)
 /* Return true if OP is valid store with update operand.  */
 
 int
-store_update_operand (op, mode)
-     rtx op;
-     enum machine_mode mode;
+store_update_operand (rtx op, enum machine_mode mode)
 {
   if (GET_CODE (op) != MEM
       || GET_MODE (op) != mode)
@@ -661,9 +670,7 @@ store_update_operand (op, mode)
    and only the standard movXX patterns are set up to handle them.  */
 
 int
-nonvol_nonimm_operand (op, mode)
-     rtx op;
-     enum machine_mode mode;
+nonvol_nonimm_operand (rtx op, enum machine_mode mode)
 {
   if (GET_CODE (op) == MEM && MEM_VOLATILE_P (op))
     return 0;
@@ -675,9 +682,7 @@ nonvol_nonimm_operand (op, mode)
    contexts.  */
 
 int
-const_sint32_operand (op, mode)
-     rtx op;
-     enum machine_mode mode ATTRIBUTE_UNUSED;
+const_sint32_operand (rtx op, enum machine_mode mode ATTRIBUTE_UNUSED)
 {
   /* All allowed constants will fit a CONST_INT.  */
   return (GET_CODE (op) == CONST_INT
@@ -689,9 +694,7 @@ const_sint32_operand (op, mode)
    need some extra crud to make it work when hosted on 64-bit machines.  */
 
 int
-const_uint32_operand (op, mode)
-     rtx op;
-     enum machine_mode mode ATTRIBUTE_UNUSED;
+const_uint32_operand (rtx op, enum machine_mode mode ATTRIBUTE_UNUSED)
 {
 #if HOST_BITS_PER_WIDE_INT > 32
   /* All allowed constants will fit a CONST_INT.  */
@@ -710,15 +713,13 @@ const_uint32_operand (op, mode)
    comparisons that use the bits that are valid.  */
 
 int
-proper_comparison_operator (op, mode)
-    rtx op;
-    enum machine_mode mode ATTRIBUTE_UNUSED;
+proper_comparison_operator (rtx op, enum machine_mode mode ATTRIBUTE_UNUSED)
 {
-  enum rtx_code code = GET_CODE (op);
-
-  if (GET_RTX_CLASS (code) != '<')
+  enum rtx_code code;
+  if (!COMPARISON_P (op))
     return 0;
 
+  code = GET_CODE (op);
   if (GET_MODE (XEXP (op, 0)) == CCZNmode)
     return (code == EQ || code == NE);
   if (GET_MODE (XEXP (op, 0)) == CCZNCmode)
@@ -733,9 +734,7 @@ proper_comparison_operator (op, mode)
    return the rtx for the cc reg in the proper mode.  */
 
 rtx
-gen_compare_reg (code, x, y)
-     enum rtx_code code;
-     rtx x, y;
+gen_compare_reg (enum rtx_code code, rtx x, rtx y)
 {
   enum machine_mode mode = SELECT_CC_MODE (code, x, y);
   rtx cc_reg;
@@ -752,8 +751,7 @@ gen_compare_reg (code, x, y)
    We assume the value can be either signed or unsigned.  */
 
 int
-arc_double_limm_p (value)
-     rtx value;
+arc_double_limm_p (rtx value)
 {
   HOST_WIDE_INT low, high;
 
@@ -788,13 +786,12 @@ arc_double_limm_p (value)
    aligned.  So we round the space up if necessary, and leave it to va_start
    to compensate.  */
 
-void
-arc_setup_incoming_varargs (cum, mode, type, pretend_size, no_rtl)
-     CUMULATIVE_ARGS *cum;
-     enum machine_mode mode;
-     tree type ATTRIBUTE_UNUSED;
-     int *pretend_size;
-     int no_rtl;
+static void
+arc_setup_incoming_varargs (CUMULATIVE_ARGS *cum,
+                            enum machine_mode mode,
+                            tree type ATTRIBUTE_UNUSED,
+                            int *pretend_size,
+                            int no_rtl)
 {
   int first_anon_arg;
 
@@ -822,9 +819,7 @@ arc_setup_incoming_varargs (cum, mode, type, pretend_size, no_rtl)
       set_mem_alias_set (regblock, get_varargs_alias_set ());
       set_mem_align (regblock, BITS_PER_WORD);
       move_block_from_reg (first_reg_offset, regblock,
-			   MAX_ARC_PARM_REGS - first_reg_offset,
-			   ((MAX_ARC_PARM_REGS - first_reg_offset)
-			    * UNITS_PER_WORD));
+			   MAX_ARC_PARM_REGS - first_reg_offset);
 
       *pretend_size = ((MAX_ARC_PARM_REGS - first_reg_offset + align_slop)
 		       * UNITS_PER_WORD);
@@ -833,18 +828,70 @@ arc_setup_incoming_varargs (cum, mode, type, pretend_size, no_rtl)
 
 /* Cost functions.  */
 
+/* Compute a (partial) cost for rtx X.  Return true if the complete
+   cost has been computed, and false if subexpressions should be
+   scanned.  In either case, *TOTAL contains the cost result.  */
+
+static bool
+arc_rtx_costs (rtx x, int code, int outer_code ATTRIBUTE_UNUSED, int *total)
+{
+  switch (code)
+    {
+      /* Small integers are as cheap as registers.  4 byte values can
+	 be fetched as immediate constants - let's give that the cost
+	 of an extra insn.  */
+    case CONST_INT:
+      if (SMALL_INT (INTVAL (x)))
+	{
+	  *total = 0;
+	  return true;
+	}
+      /* FALLTHRU */
+
+    case CONST:
+    case LABEL_REF:
+    case SYMBOL_REF:
+      *total = COSTS_N_INSNS (1);
+      return true;
+
+    case CONST_DOUBLE:
+      {
+        rtx high, low;
+        split_double (x, &high, &low);
+	*total = COSTS_N_INSNS (!SMALL_INT (INTVAL (high))
+				+ !SMALL_INT (INTVAL (low)));
+	return true;
+      }
+
+    /* Encourage synth_mult to find a synthetic multiply when reasonable.
+       If we need more than 12 insns to do a multiply, then go out-of-line,
+       since the call overhead will be < 10% of the cost of the multiply.  */
+    case ASHIFT:
+    case ASHIFTRT:
+    case LSHIFTRT:
+      if (TARGET_SHIFTER)
+        *total = COSTS_N_INSNS (1);
+      else if (GET_CODE (XEXP (x, 1)) != CONST_INT)
+        *total = COSTS_N_INSNS (16);
+      else
+        *total = COSTS_N_INSNS (INTVAL (XEXP ((x), 1)));
+      return false;
+
+    default:
+      return false;
+    }
+}
+
+
 /* Provide the costs of an addressing mode that contains ADDR.
    If ADDR is not a valid address, its cost is irrelevant.  */
 
-int
-arc_address_cost (addr)
-     rtx addr;
+static int
+arc_address_cost (rtx addr)
 {
   switch (GET_CODE (addr))
     {
     case REG :
-      /* This is handled in the macro that calls us.
-	 It's here for documentation.  */
       return 1;
 
     case LABEL_REF :
@@ -955,8 +1002,7 @@ static struct arc_frame_info zero_frame_info;
    call with DECL = NULL_TREE.  */
 
 enum arc_function_type
-arc_compute_function_type (decl)
-     tree decl;
+arc_compute_function_type (tree decl)
 {
   tree a;
   /* Cached value.  */
@@ -1026,8 +1072,7 @@ arc_compute_function_type (decl)
    SIZE is the size needed for local variables.  */
 
 unsigned int
-arc_compute_frame_size (size)
-     int size;			/* # of var. bytes allocated.  */
+arc_compute_frame_size (int size /* # of var. bytes allocated.  */)
 {
   int regno;
   unsigned int total_size, var_size, args_size, pretend_size, extra_size;
@@ -1091,12 +1136,11 @@ arc_compute_frame_size (size)
 /* Common code to save/restore registers.  */
 
 void
-arc_save_restore (file, base_reg, offset, gmask, op)
-     FILE *file;
-     const char *base_reg;
-     unsigned int offset;
-     unsigned int gmask;
-     const char *op;
+arc_save_restore (FILE *file,
+                  const char *base_reg,
+                  unsigned int offset,
+                  unsigned int gmask,
+                  const char *op)
 {
   int regno;
 
@@ -1119,13 +1163,10 @@ arc_save_restore (file, base_reg, offset, gmask, op)
    symbols.  */
 
 static bool
-arc_assemble_integer (x, size, aligned_p)
-     rtx x;
-     unsigned int size;
-     int aligned_p;
+arc_assemble_integer (rtx x, unsigned int size, int aligned_p)
 {
   if (size == UNITS_PER_WORD && aligned_p
-      && ((GET_CODE (x) == SYMBOL_REF && SYMBOL_REF_FLAG (x))
+      && ((GET_CODE (x) == SYMBOL_REF && SYMBOL_REF_FUNCTION_P (x))
 	  || GET_CODE (x) == LABEL_REF))
     {
       fputs ("\t.word\t%st(", asm_out_file);
@@ -1139,9 +1180,7 @@ arc_assemble_integer (x, size, aligned_p)
 /* Set up the stack and frame pointer (if desired) for the function.  */
 
 static void
-arc_output_function_prologue (file, size)
-     FILE *file;
-     HOST_WIDE_INT size;
+arc_output_function_prologue (FILE *file, HOST_WIDE_INT size)
 {
   const char *sp_str = reg_names[STACK_POINTER_REGNUM];
   const char *fp_str = reg_names[FRAME_POINTER_REGNUM];
@@ -1199,7 +1238,7 @@ arc_output_function_prologue (file, size)
 
   /* Allocate the stack frame.  */
   if (size - current_frame_info.pretend_size > 0)
-    fprintf (file, "\tsub %s,%s,%d\n",
+    fprintf (file, "\tsub %s,%s," HOST_WIDE_INT_PRINT_DEC "\n",
 	     sp_str, sp_str, size - current_frame_info.pretend_size);
 
   /* Save any needed call-saved regs (and call-used if this is an
@@ -1217,9 +1256,7 @@ arc_output_function_prologue (file, size)
    and regs.  */
 
 static void
-arc_output_function_epilogue (file, size)
-     FILE *file;
-     HOST_WIDE_INT size;
+arc_output_function_epilogue (FILE *file, HOST_WIDE_INT size)
 {
   rtx epilogue_delay = current_function_epilogue_delay_list;
   int noepilogue = FALSE;
@@ -1291,7 +1328,7 @@ arc_output_function_epilogue (file, size)
       /* ??? If stack intactness is important, always emit now.  */
       if (MUST_SAVE_RETURN_ADDR && epilogue_delay != NULL_RTX)
 	{
-	  final_scan_insn (XEXP (epilogue_delay, 0), file, 1, -2, 1);
+	  final_scan_insn (XEXP (epilogue_delay, 0), file, 1, -2, 1, NULL);
 	  epilogue_delay = NULL_RTX;
 	}
 
@@ -1323,7 +1360,8 @@ arc_output_function_epilogue (file, size)
 	{
 	  if (epilogue_delay)
 	    {
-	      final_scan_insn (XEXP (epilogue_delay, 0), file, 1, -2, 1);
+	      final_scan_insn (XEXP (epilogue_delay, 0), file, 1, -2, 1,
+			       NULL);
 	    }
 	}
 
@@ -1348,7 +1386,7 @@ arc_output_function_epilogue (file, size)
 	    abort ();
 	  if (restored < size)
 	    abort ();
-	  final_scan_insn (XEXP (epilogue_delay, 0), file, 1, -2, 1);
+	  final_scan_insn (XEXP (epilogue_delay, 0), file, 1, -2, 1, NULL);
 	}
       else if (frame_pointer_needed && !fp_restored_p)
 	{
@@ -1361,7 +1399,7 @@ arc_output_function_epilogue (file, size)
 	{
 	  if (!SMALL_INT (size - restored))
 	    abort ();
-	  fprintf (file, "\tadd %s,%s,%d\n",
+	  fprintf (file, "\tadd %s,%s," HOST_WIDE_INT_PRINT_DEC "\n",
 		   sp_str, sp_str, size - restored);
 	}
       else
@@ -1382,7 +1420,7 @@ arc_output_function_epilogue (file, size)
    delay slot so for now we only consider functions with empty frames.  */
 
 int
-arc_delay_slots_for_epilogue ()
+arc_delay_slots_for_epilogue (void)
 {
   if (arc_compute_function_type (current_function_decl) != ARC_FUNCTION_NORMAL)
     return 0;
@@ -1398,9 +1436,7 @@ arc_delay_slots_for_epilogue ()
    pointer or any call-saved register is OK.  SLOT will always be 0.  */
 
 int
-arc_eligible_for_epilogue_delay (trial, slot)
-     rtx trial;
-     int slot;
+arc_eligible_for_epilogue_delay (rtx trial, int slot)
 {
   if (slot != 0)
     abort ();
@@ -1425,7 +1461,7 @@ arc_eligible_for_epilogue_delay (trial, slot)
 /* Emit special PIC prologues and epilogues.  */
 
 void
-arc_finalize_pic ()
+arc_finalize_pic (void)
 {
   /* nothing to do */
 }
@@ -1433,9 +1469,7 @@ arc_finalize_pic ()
 /* Return true if OP is a shift operator.  */
 
 int
-shift_operator (op, mode)
-     rtx op;
-     enum machine_mode mode ATTRIBUTE_UNUSED;
+shift_operator (rtx op, enum machine_mode mode ATTRIBUTE_UNUSED)
 {
   switch (GET_CODE (op))
     {
@@ -1463,8 +1497,7 @@ shift_operator (op, mode)
    using it here will give us a chance to play with it.  */
 
 const char *
-output_shift (operands)
-     rtx *operands;
+output_shift (rtx *operands)
 {
   rtx shift = operands[3];
   enum machine_mode mode = GET_MODE (shift);
@@ -1585,19 +1618,20 @@ output_shift (operands)
    CXT is an RTX for the static chain value for the function.  */
 
 void
-arc_initialize_trampoline (tramp, fnaddr, cxt)
-     rtx tramp ATTRIBUTE_UNUSED, fnaddr ATTRIBUTE_UNUSED, cxt ATTRIBUTE_UNUSED;
+arc_initialize_trampoline (rtx tramp ATTRIBUTE_UNUSED,
+                           rtx fnaddr ATTRIBUTE_UNUSED,
+                           rtx cxt ATTRIBUTE_UNUSED)
 {
 }
 
 /* Set the cpu type and print out other fancy things,
    at the top of the file.  */
 
-void
-arc_asm_file_start (file)
-     FILE *file;
+static void
+arc_file_start (void)
 {
-  fprintf (file, "\t.cpu %s\n", arc_cpu_string);
+  default_file_start ();
+  fprintf (asm_out_file, "\t.cpu %s\n", arc_cpu_string);
 }
 
 /* Print operand X (an rtx) in assembler syntax to file FILE.
@@ -1605,10 +1639,7 @@ arc_asm_file_start (file)
    For `%' followed by punctuation, CODE is the punctuation and X is null.  */
 
 void
-arc_print_operand (file, x, code)
-     FILE *file;
-     rtx x;
-     int code;
+arc_print_operand (FILE *file, rtx x, int code)
 {
   switch (code)
     {
@@ -1708,7 +1739,7 @@ arc_print_operand (file, x, code)
 	output_operand_lossage ("invalid operand to %%R code");
       return;
     case 'S' :
-      if ((GET_CODE (x) == SYMBOL_REF && SYMBOL_REF_FLAG (x))
+      if ((GET_CODE (x) == SYMBOL_REF && SYMBOL_REF_FUNCTION_P (x))
 	  || GET_CODE (x) == LABEL_REF)
 	{
 	  fprintf (file, "%%st(");
@@ -1820,9 +1851,7 @@ arc_print_operand (file, x, code)
 /* Print a memory address as an operand to reference that memory location.  */
 
 void
-arc_print_operand_address (file, addr)
-     FILE *file;
-     rtx addr;
+arc_print_operand_address (FILE *file, rtx addr)
 {
   register rtx base, index = 0;
   int offset = 0;
@@ -1833,7 +1862,7 @@ arc_print_operand_address (file, addr)
       fputs (reg_names[REGNO (addr)], file);
       break;
     case SYMBOL_REF :
-      if (/*???*/ 0 && SYMBOL_REF_FLAG (addr))
+      if (/*???*/ 0 && SYMBOL_REF_FUNCTION_P (addr))
 	{
 	  fprintf (file, "%%st(");
 	  output_addr_const (file, addr);
@@ -1879,8 +1908,7 @@ arc_print_operand_address (file, addr)
 /* Update compare/branch separation marker.  */
 
 static void
-record_cc_ref (insn)
-     rtx insn;
+record_cc_ref (rtx insn)
 {
   last_insn_set_cc_p = current_insn_set_cc_p;
 
@@ -1922,7 +1950,7 @@ record_cc_ref (insn)
    0 -> 2 final_prescan_insn, if the `target' is an unconditional branch
    1 -> 3 branch patterns, after having not output the conditional branch
    2 -> 4 branch patterns, after having not output the conditional branch
-   3 -> 0 ASM_OUTPUT_INTERNAL_LABEL, if the `target' label is reached
+   3 -> 0 (*targetm.asm_out.internal_label), if the `target' label is reached
           (the target label has CODE_LABEL_NUMBER equal to
 	  arc_ccfsm_target_label).
    4 -> 0 final_prescan_insn, if `target' unconditional branch is reached
@@ -1936,15 +1964,14 @@ record_cc_ref (insn)
    we may not be outputting the branch.  */
 
 void
-arc_final_prescan_insn (insn, opvec, noperands)
-     rtx insn;
-     rtx *opvec ATTRIBUTE_UNUSED;
-     int noperands ATTRIBUTE_UNUSED;
+arc_final_prescan_insn (rtx insn,
+                        rtx *opvec ATTRIBUTE_UNUSED,
+                        int noperands ATTRIBUTE_UNUSED)
 {
   /* BODY will hold the body of INSN.  */
   register rtx body = PATTERN (insn);
 
-  /* This will be 1 if trying to repeat the trick (ie: do the `else' part of
+  /* This will be 1 if trying to repeat the trick (i.e.: do the `else' part of
      an if/then/else), and things need to be reversed.  */
   int reverse = 0;
 
@@ -2218,12 +2245,10 @@ arc_final_prescan_insn (insn, opvec, noperands)
 /* Record that we are currently outputting label NUM with prefix PREFIX.
    It it's the label we're looking for, reset the ccfsm machinery.
 
-   Called from ASM_OUTPUT_INTERNAL_LABEL.  */
+   Called from (*targetm.asm_out.internal_label).  */
 
 void
-arc_ccfsm_at_label (prefix, num)
-     const char *prefix;
-     int num;
+arc_ccfsm_at_label (const char *prefix, int num)
 {
   if (arc_ccfsm_state == 3 && arc_ccfsm_target_label == num
       && !strcmp (prefix, "L"))
@@ -2237,7 +2262,7 @@ arc_ccfsm_at_label (prefix, num)
    deleted.  */
 
 int
-arc_ccfsm_branch_deleted_p ()
+arc_ccfsm_branch_deleted_p (void)
 {
   if (arc_ccfsm_state == 1 || arc_ccfsm_state == 2)
     return 1;
@@ -2248,7 +2273,7 @@ arc_ccfsm_branch_deleted_p ()
    conditionalized.  */
 
 void
-arc_ccfsm_record_branch_deleted ()
+arc_ccfsm_record_branch_deleted (void)
 {
   /* Indicate we're conditionalizing insns now.  */
   arc_ccfsm_state += 2;
@@ -2260,9 +2285,7 @@ arc_ccfsm_record_branch_deleted ()
 }
 
 void
-arc_va_start (valist, nextarg)
-     tree valist;
-     rtx nextarg;
+arc_va_start (tree valist, rtx nextarg)
 {
   /* See arc_setup_incoming_varargs for reasons for this oddity.  */
   if (current_function_args_info < 8
@@ -2272,89 +2295,66 @@ arc_va_start (valist, nextarg)
   std_expand_builtin_va_start (valist, nextarg);
 }
 
-rtx
-arc_va_arg (valist, type)
-     tree valist, type;
-{
-  rtx addr_rtx;
-  tree addr, incr;
-  tree type_ptr = build_pointer_type (type);
-
-  /* All aggregates are passed by reference.  All scalar types larger
-     than 8 bytes are passed by reference.  */
-
-  if (AGGREGATE_TYPE_P (type) || int_size_in_bytes (type) > 8)
-    {
-      tree type_ptr_ptr = build_pointer_type (type_ptr);
-
-      addr = build (INDIRECT_REF, type_ptr,
-		    build (NOP_EXPR, type_ptr_ptr, valist));
-
-      incr = build (PLUS_EXPR, TREE_TYPE (valist),
-		    valist, build_int_2 (UNITS_PER_WORD, 0));
-    }
-  else
-    {
-      HOST_WIDE_INT align, rounded_size;
-
-      /* Compute the rounded size of the type.  */
-      align = PARM_BOUNDARY / BITS_PER_UNIT;
-      rounded_size = (((TREE_INT_CST_LOW (TYPE_SIZE (type)) / BITS_PER_UNIT
-			+ align - 1) / align) * align);
-
-      /* Align 8 byte operands.  */
-      addr = valist;
-      if (TYPE_ALIGN (type) > BITS_PER_WORD)
-	{
-	  /* AP = (TYPE *)(((int)AP + 7) & -8)  */
-
-	  addr = build (NOP_EXPR, integer_type_node, valist);
-	  addr = fold (build (PLUS_EXPR, integer_type_node, addr,
-			      build_int_2 (7, 0)));
-	  addr = fold (build (BIT_AND_EXPR, integer_type_node, addr,
-			      build_int_2 (-8, 0)));
-	  addr = fold (build (NOP_EXPR, TREE_TYPE (valist), addr));
-	}
-
-      /* The increment is always rounded_size past the aligned pointer.  */
-      incr = fold (build (PLUS_EXPR, TREE_TYPE (addr), addr,
-			  build_int_2 (rounded_size, 0)));
-
-      /* Adjust the pointer in big-endian mode.  */
-      if (BYTES_BIG_ENDIAN)
-	{
-	  HOST_WIDE_INT adj;
-	  adj = TREE_INT_CST_LOW (TYPE_SIZE (type)) / BITS_PER_UNIT;
-	  if (rounded_size > align)
-	    adj = rounded_size;
-
-	  addr = fold (build (PLUS_EXPR, TREE_TYPE (addr), addr,
-			      build_int_2 (rounded_size - adj, 0)));
-	}
-    }
-
-  /* Evaluate the data address.  */
-  addr_rtx = expand_expr (addr, NULL_RTX, Pmode, EXPAND_NORMAL);
-  addr_rtx = copy_to_reg (addr_rtx);
-  
-  /* Compute new value for AP.  */
-  incr = build (MODIFY_EXPR, TREE_TYPE (valist), valist, incr);
-  TREE_SIDE_EFFECTS (incr) = 1;
-  expand_expr (incr, const0_rtx, VOIDmode, EXPAND_NORMAL);
-
-  return addr_rtx;
-}
-
-/* On the ARC, function addresses are not the same as normal addresses.
-   Branch to absolute address insns take an address that is right-shifted
-   by 2.  We encode the fact that we have a function here, and then emit a
-   special assembler op when outputting the address.  */
+/* This is how to output a definition of an internal numbered label where
+   PREFIX is the class of label and NUM is the number within the class.  */
 
 static void
-arc_encode_section_info (decl, first)
-     tree decl;
-     int first ATTRIBUTE_UNUSED;
+arc_internal_label (FILE *stream, const char *prefix, unsigned long labelno)
 {
-  if (TREE_CODE (decl) == FUNCTION_DECL)
-    SYMBOL_REF_FLAG (XEXP (DECL_RTL (decl), 0)) = 1;
+  arc_ccfsm_at_label (prefix, labelno);
+  default_internal_label (stream, prefix, labelno);
+}
+
+/* Worker function for TARGET_ASM_EXTERNAL_LIBCALL.  */
+
+static void
+arc_external_libcall (rtx fun ATTRIBUTE_UNUSED)
+{
+#if 0
+/* On the ARC we want to have libgcc's for multiple cpus in one binary.
+   We can't use `assemble_name' here as that will call ASM_OUTPUT_LABELREF
+   and we'll get another suffix added on if -mmangle-cpu.  */
+  if (TARGET_MANGLE_CPU_LIBGCC)
+    {
+      fprintf (FILE, "\t.rename\t_%s, _%s%s\n",
+	       XSTR (SYMREF, 0), XSTR (SYMREF, 0),
+	       arc_mangle_suffix);
+    }
+#endif
+}
+
+/* Worker function for TARGET_RETURN_IN_MEMORY.  */
+
+static bool
+arc_return_in_memory (tree type, tree fntype ATTRIBUTE_UNUSED)
+{
+  if (AGGREGATE_TYPE_P (type))
+    return true;
+  else
+    {
+      HOST_WIDE_INT size = int_size_in_bytes (type);
+      return (size == -1 || size > 8);
+    }
+}
+
+/* For ARC, All aggregates and arguments greater than 8 bytes are
+   passed by reference.  */
+
+static bool
+arc_pass_by_reference (CUMULATIVE_ARGS *ca ATTRIBUTE_UNUSED,
+		       enum machine_mode mode, tree type,
+		       bool named ATTRIBUTE_UNUSED)
+{
+  unsigned HOST_WIDE_INT size;
+
+  if (type)
+    {
+      if (AGGREGATE_TYPE_P (type))
+	return true;
+      size = int_size_in_bytes (type);
+    }
+  else
+    size = GET_MODE_SIZE (mode);
+
+  return size > 8;
 }
