@@ -1,5 +1,5 @@
 /* Subroutines for assembler code output on the NS32000.
-   Copyright (C) 1988, 1994, 1995, 1996, 1997, 1998, 1999, 2000, 2001
+   Copyright (C) 1988, 1994, 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2004
    Free Software Foundation, Inc.
 
 This file is part of GCC.
@@ -73,6 +73,9 @@ static void ns32k_output_function_prologue (FILE *, HOST_WIDE_INT);
 static void ns32k_output_function_epilogue (FILE *, HOST_WIDE_INT);
 static bool ns32k_rtx_costs (rtx, int, int, int *);
 static int ns32k_address_cost (rtx);
+static rtx ns32k_struct_value_rtx (tree, int);
+static int ns32k_arg_partial_bytes (CUMULATIVE_ARGS *, enum machine_mode,
+				    tree, bool);
 
 /* Initialize the GCC target structure.  */
 #undef TARGET_ATTRIBUTE_TABLE
@@ -96,6 +99,12 @@ static int ns32k_address_cost (rtx);
 #undef TARGET_ADDRESS_COST
 #define TARGET_ADDRESS_COST ns32k_address_cost
 
+#undef TARGET_STRUCT_VALUE_RTX
+#define TARGET_STRUCT_VALUE_RTX ns32k_struct_value_rtx
+
+#undef TARGET_ARG_PARTIAL_BYTES
+#define TARGET_ARG_PARTIAL_BYTES ns32k_arg_partial_bytes
+
 #undef TARGET_ASM_FILE_START_APP_OFF
 #define TARGET_ASM_FILE_START_APP_OFF true
 
@@ -539,7 +548,7 @@ register_move_cost (enum reg_class CLASS1, enum reg_class CLASS2)
 
 #if 0
 /* We made the insn definitions copy from floating point to general
-  registers via the stack. */
+  registers via the stack.  */
 int
 secondary_memory_needed (enum reg_class CLASS1,
 			 enum reg_class CLASS2,
@@ -554,7 +563,7 @@ secondary_memory_needed (enum reg_class CLASS1,
 
 /* TARGET_ADDRESS_COST calls this.  This function is not optimal
    for the 32032 & 32332, but it probably is better than
-   the default. */
+   the default.  */
 
 static int
 ns32k_address_cost (rtx operand)
@@ -878,12 +887,12 @@ expand_block_move (rtx operands[])
 	      /* Use movmd. It is slower than multiple movd's but more
 		 compact. It is also slower than movsd for large copies
 		 but causes less registers reloading so is better than movsd
-		 for small copies. */
+		 for small copies.  */
 	      rtx src, dest;
 	      dest = copy_addr_to_reg (XEXP (operands[0], 0));
 	      src = copy_addr_to_reg (XEXP (operands[1], 0));
 	    
-	      emit_insn (gen_movstrsi2(dest, src, GEN_INT (words)));
+	      emit_insn (gen_movmemsi2(dest, src, GEN_INT (words)));
 	    }
 	}
       move_tail (operands, bytes & 3, bytes & ~3);
@@ -910,7 +919,7 @@ expand_block_move (rtx operands[])
       if (bytes >> 2)
 	{
 	  emit_move_insn (count_reg, GEN_INT (bytes >> 2));
-	  emit_insn (gen_movstrsi1 (GEN_INT (4)));
+	  emit_insn (gen_movmemsi1 (GEN_INT (4)));
 	}
       /* insns to copy rest */
       move_tail (operands, bytes & 3, 0);
@@ -918,8 +927,8 @@ expand_block_move (rtx operands[])
   else if (align == UNITS_PER_WORD)
     {
       /* insns to copy by words */
-      emit_insn (gen_lshrsi3 (count_reg, bytes_rtx, GEN_INT (2)));
-      emit_insn (gen_movstrsi1 (GEN_INT (4)));
+      emit_insn (gen_lshrsi3 (count_reg, bytes_rtx, const2_rtx));
+      emit_insn (gen_movmemsi1 (GEN_INT (4)));
       if (constp)
 	{
 	  move_tail (operands, bytes & 3, 0);
@@ -928,7 +937,7 @@ expand_block_move (rtx operands[])
 	{
 	  /* insns to copy rest */
 	  emit_insn (gen_andsi3 (count_reg, bytes_rtx, GEN_INT (3)));
-	  emit_insn (gen_movstrsi1 (const1_rtx));
+	  emit_insn (gen_movmemsi1 (const1_rtx));
 	}
     }
   else
@@ -954,17 +963,17 @@ expand_block_move (rtx operands[])
       emit_insn (gen_negsi2 (count_reg, src_reg));
       emit_insn (gen_andsi3 (count_reg, count_reg, GEN_INT (3)));
       emit_insn (gen_subsi3 (bytes_reg, bytes_reg, count_reg));
-      emit_insn (gen_movstrsi1 (const1_rtx));
+      emit_insn (gen_movmemsi1 (const1_rtx));
       if (!constp)
 	emit_label (aligned_label);
 
       /* insns to copy by words */
-      emit_insn (gen_lshrsi3 (count_reg, bytes_reg, GEN_INT (2)));
-      emit_insn (gen_movstrsi1 (GEN_INT (4)));
+      emit_insn (gen_lshrsi3 (count_reg, bytes_reg, const2_rtx));
+      emit_insn (gen_movmemsi1 (GEN_INT (4)));
 
       /* insns to copy rest */
       emit_insn (gen_andsi3 (count_reg, bytes_reg, GEN_INT (3)));
-      emit_insn (gen_movstrsi1 (const1_rtx));
+      emit_insn (gen_movmemsi1 (const1_rtx));
     }
 }
 
@@ -1061,7 +1070,7 @@ ns32k_handle_fntype_attribute (tree *node, tree name,
       && TREE_CODE (*node) != FIELD_DECL
       && TREE_CODE (*node) != TYPE_DECL)
     {
-      warning ("`%s' attribute only applies to functions",
+      warning ("%qs attribute only applies to functions",
 	       IDENTIFIER_POINTER (name));
       *no_add_attrs = true;
     }
@@ -1557,3 +1566,105 @@ output_move_dconst (int n, const char *s)
   strcat (r, s);
   return r;
 }
+
+static rtx
+ns32k_struct_value_rtx (tree fntype ATTRIBUTE_UNUSED,
+			int incoming ATTRIBUTE_UNUSED)
+{
+  return gen_rtx_REG (Pmode, NS32K_STRUCT_VALUE_REGNUM);
+}
+
+/* Worker function for NOTICE_UPDATE_CC.  */
+
+void
+ns32k_notice_update_cc (rtx exp, rtx insn ATTRIBUTE_UNUSED)
+{
+  if (GET_CODE (exp) == SET)
+    {
+      if (GET_CODE (SET_DEST (exp)) == CC0)
+	{
+	  cc_status.flags = 0;
+	  cc_status.value1 = SET_DEST (exp);
+	  cc_status.value2 = SET_SRC (exp);
+	}
+      else if (GET_CODE (SET_SRC (exp)) == CALL)
+	{
+	  CC_STATUS_INIT;
+	}
+      else if (GET_CODE (SET_DEST (exp)) == REG)
+	{
+	  if (cc_status.value1
+	      && reg_overlap_mentioned_p (SET_DEST (exp), cc_status.value1))
+	    cc_status.value1 = 0;
+	  if (cc_status.value2
+	      && reg_overlap_mentioned_p (SET_DEST (exp), cc_status.value2))
+	    cc_status.value2 = 0;
+	}
+      else if (GET_CODE (SET_DEST (exp)) == MEM)
+	{
+	  CC_STATUS_INIT;
+	}
+    }
+  else if (GET_CODE (exp) == PARALLEL
+	   && GET_CODE (XVECEXP (exp, 0, 0)) == SET)
+    {
+      if (GET_CODE (SET_DEST (XVECEXP (exp, 0, 0))) == CC0)
+	{
+	  cc_status.flags = 0;
+	  cc_status.value1 = SET_DEST (XVECEXP (exp, 0, 0));
+	  cc_status.value2 = SET_SRC (XVECEXP (exp, 0, 0));
+	}
+      else if (GET_CODE (SET_DEST (XVECEXP (exp, 0, 0))) == REG)
+	{
+	  if (cc_status.value1
+	      && reg_overlap_mentioned_p (SET_DEST (XVECEXP (exp, 0, 0)),
+					  cc_status.value1))
+	    cc_status.value1 = 0;
+	  if (cc_status.value2
+	      && reg_overlap_mentioned_p (SET_DEST (XVECEXP (exp, 0, 0)),
+					  cc_status.value2))
+	    cc_status.value2 = 0;
+	}
+      else if (GET_CODE (SET_DEST (XVECEXP (exp, 0, 0))) == MEM)
+	{
+	  CC_STATUS_INIT;
+	}
+    }
+  else if (GET_CODE (exp) == CALL)
+    {
+      /* all bets are off */
+      CC_STATUS_INIT;
+    }
+  else
+    {
+      /* nothing happens? CC_STATUS_INIT; */
+    }
+  if (cc_status.value1 && GET_CODE (cc_status.value1) == REG
+      && cc_status.value2
+      && reg_overlap_mentioned_p (cc_status.value1, cc_status.value2))
+    abort ();
+}
+
+/* Implement TARGET_ARG_PARTIAL_BYTES.  */
+
+static int
+ns32k_arg_partial_bytes (CUMULATIVE_ARGS *pcum, enum machine_mode mode,
+			 tree type, bool named ATTRIBUTE_UNUSED)
+{
+  int cum = *pcum;
+
+  if (TARGET_REGPARM && cum < 8)
+    {
+      HOST_WIDE_INT size;
+
+      if (mode == BLKmode)
+	size = int_size_in_bytes (type);
+      else
+	size = GET_MODE_SIZE (mode);
+
+      if (8 < cum + size)
+	return 8 - cum;
+    }
+
+  return 0;
+}
